1#ifndef AMREX_EB2_LEVEL_H_
2#define AMREX_EB2_LEVEL_H_
3#include <AMReX_Config.H>
14#if (AMREX_SPACEDIM == 3)
18#include <AMReX_EB2_C.H>
29#include <unordered_map>
39template <
typename G>
class GShopLevel;
134 void write_to_chkpt_file (
const std::string& fname,
bool extend_domain_face,
int max_grid_size)
const;
144 void setShift (
int direction,
int ncells);
186 if (idim != face_dir) { edge_type[idim] = 0; }
191 [[nodiscard]]
bool hasFCData (
int face_dir)
const noexcept {
192 return face_dir >= 0 && face_dir < AMREX_SPACEDIM &&
250#if (AMREX_SPACEDIM == 3)
278 void setRegularLevel () {
294 int ngrow,
bool extend_domain_face,
int num_crse_opt);
304 int max_grid_size,
int ngrow,
bool extend_domain_face,
int num_crse_opt);
306#if (AMREX_SPACEDIM == 3)
310 template <
typename GS = G>
311 requires (std::same_as<GS,STLtools>)
313 int max_grid_size,
int ngrow,
bool extend_domain_face,
int num_crse_opt);
328 void buildFCData (G
const& gshop,
int face_dir,
int max_grid_size);
334 void prepare_grids (G
const& gshop,
Geometry const& geom,
int max_grid_size,
int ngrow,
335 bool extend_domain_face,
int num_crse_opt);
346 int max_grid_size,
int ngrow,
bool extend_domain_face,
int num_crse_opt)
349 if (std::is_same_v<typename G::FunctionType, AllRegularIF>) {
355 define_fine(gshop, geom, max_grid_size, ngrow, extend_domain_face, num_crse_opt);
361 int max_grid_size,
int ngrow,
bool extend_domain_face,
int num_crse_opt)
364 amrex::Print() <<
"AMReX WARNING: extend_domain_face=false is not recommended!\n";
367 BL_PROFILE(
"EB2::GShopLevel()-prepare_grids");
370 m_ngrow =
IntVect{
static_cast<int>(std::ceil(ngrow/16.)) * 16};
373 Box domain_grown = domain;
374 for (
int idim = 0; idim < AMREX_SPACEDIM; ++idim) {
378 m_ngrow[idim] = std::min(m_ngrow[idim], domain_grown.length(idim));
381 domain_grown.grow(m_ngrow);
382 m_bounding_box = (extend_domain_face) ? domain : domain_grown;
383 m_bounding_box.surroundingNodes();
386 BoxList covered_boxes;
391 num_crse_opt = std::max(0,std::min(8,num_crse_opt));
392 for (
int clev = num_crse_opt; clev >= 0; --clev) {
394 if (domain.coarsenable(crse_ratio)) {
398 if (cut_boxes.isEmpty()) {
399 covered_boxes.clear();
400 test_boxes = BoxList(crse_geom.Domain());
401 test_boxes.maxSize(max_grid_size);
403 test_boxes.swap(cut_boxes);
404 test_boxes.coarsen(crse_ratio);
405 test_boxes.maxSize(max_grid_size);
408 const Long nboxes = test_boxes.size();
409 const auto& boxes = test_boxes.data();
410 for (
Long i = iproc; i < nboxes; i += nprocs) {
411 const Box& vbx = boxes[i];
413 auto box_type = gshop.getBoxType(gbx&crse_bounding_box,crse_geom,
RunOn::Gpu);
414 if (box_type == gshop.allcovered) {
416 }
else if (box_type == gshop.mixedcells) {
428 auto grow_at_domain_boundary = [&] (BoxList& bl)
430 const IntVect& domlo = domain.smallEnd();
431 const IntVect& domhi = domain.bigEnd();
433 for (
int idim = 0; idim < AMREX_SPACEDIM; ++idim) {
434 if (m_ngrow[idim] != 0) {
435 if (b.smallEnd(idim) == domlo[idim]) {
436 b.growLo(idim,m_ngrow[idim]);
438 if (b.bigEnd(idim) == domhi[idim]) {
439 b.growHi(idim,m_ngrow[idim]);
445 grow_at_domain_boundary(covered_boxes);
446 grow_at_domain_boundary(cut_boxes);
449 if ( cut_boxes.isEmpty() &&
450 !covered_boxes.isEmpty())
452 amrex::Abort(
"AMReX_EB2_Level.H: Domain is completely covered");
455 if (!covered_boxes.isEmpty()) {
456 if (num_crse_opt > 2) {
457 covered_boxes.maxSize(max_grid_size*4);
459 m_covered_grids = BoxArray(std::move(covered_boxes));
462 if (cut_boxes.isEmpty()) {
463 m_grids = BoxArray();
464 m_dmap = DistributionMapping();
468 m_grids = BoxArray(std::move(cut_boxes));
469 m_dmap = DistributionMapping(m_grids);
476 int max_grid_size,
int ngrow,
bool extend_domain_face,
int num_crse_opt)
480#ifdef AMREX_USE_FLOAT
481 Real small_volfrac = 1.e-5_rt;
483 Real small_volfrac = 1.e-14;
485 bool cover_multiple_cuts =
false;
490 pp.
queryAdd(
"cover_multiple_cuts", cover_multiple_cuts);
493 maxiter = std::min(100000, maxiter);
495 prepare_grids(gshop, geom, max_grid_size, ngrow, extend_domain_face, num_crse_opt);
497 if (m_allregular) {
return; }
499 m_mgf.define(m_grids, m_dmap);
502 mf_info.
SetTag(
"EB2::Level");
503 m_cellflag.define(m_grids, m_dmap, 1, ng, mf_info);
504 m_volfrac.define(m_grids, m_dmap, 1, ng, mf_info);
505 m_centroid.define(m_grids, m_dmap, AMREX_SPACEDIM, ng, mf_info);
506 m_bndryarea.define(m_grids, m_dmap, 1, ng, mf_info);
507 m_bndrycent.define(m_grids, m_dmap, AMREX_SPACEDIM, ng, mf_info);
508 m_bndrynorm.define(m_grids, m_dmap, AMREX_SPACEDIM, ng, mf_info);
509 for (
int idim = 0; idim < AMREX_SPACEDIM; ++idim) {
511 m_dmap, 1, ng, mf_info);
513 m_dmap, AMREX_SPACEDIM-1, ng, mf_info);
514 IntVect edge_type{1}; edge_type[idim] = 0;
515 m_edgecent[idim].define(
amrex::convert(m_grids, edge_type), m_dmap, 1, ng, mf_info);
521 auto bounding_box = m_bounding_box;
522 for (
int idim = 0; idim < AMREX_SPACEDIM; ++idim) {
523 if (!extend_domain_face || geom.
isPeriodic(idim)) {
535 for (; iter < maxiter; ++iter)
540#pragma omp parallel if (Gpu::notInLaunchRegion()) reduction(+:nsmallcells,nmulticuts)
543#if (AMREX_SPACEDIM == 3)
549 auto& gfab = m_mgf[mfi];
550 const Box& vbx = gfab.validbox();
552 auto&
levelset = gfab.getLevelSet();
554 gshop.fillFab(
levelset, geom, gshop_run_on, bounding_box);
581 auto& facetype = gfab.getFaceType();
589#if (AMREX_SPACEDIM == 3)
590 auto& edgetype = gfab.getEdgeType();
603 for (
int idim = 0; idim < AMREX_SPACEDIM; ++idim) {
604 edgetype[idim].prefetchToHost();
605 m_edgecent[idim][mfi].prefetchToHost();
610 gshop.getIntercept({xip,yip,zip}, {xdg,ydg,zdg}, clst,
611 geom, gshop_run_on, bounding_box);
615 for (
int idim = 0; idim < AMREX_SPACEDIM; ++idim) {
616 edgetype[idim].prefetchToDevice();
617 m_edgecent[idim][mfi].prefetchToDevice();
622 gshop.updateIntercept({xip,yip,zip}, {xdg,ydg,zdg}, clst, geom);
625 for (
int idim = 0; idim < AMREX_SPACEDIM; ++idim) {
626 const Box& b = facetype[idim].box();
633 nmc = build_faces(vbx, cfg, ftx, fty, ftz, xdg, ydg, zdg, lst,
635 xm2, ym2, zm2, dx, problo, cover_multiple_cuts);
637 cellflagtmp.
resize(m_cellflag[mfi].box());
641 build_cells(vbx, cfg, ftx, fty, ftz,
apx,
apy,
apz,
642 fcx,
fcy,
fcz, xm2, ym2, zm2, dx, vfr, ctr,
643 bar, bct, bnm, cfgtmp, lst,
644 small_volfrac, geom, extend_domain_face, cover_multiple_cuts,
652#elif (AMREX_SPACEDIM == 2)
660 for (
int idim = 0; idim < AMREX_SPACEDIM; ++idim) {
661 facetype[idim].prefetchToHost();
662 m_edgecent[idim][mfi].prefetchToHost();
668 gshop.getIntercept({xip,yip},
669 {facetype[1].const_array(), facetype[0].const_array()},
670 clst, geom, gshop_run_on, bounding_box);
674 for (
int idim = 0; idim < AMREX_SPACEDIM; ++idim) {
675 facetype[idim].prefetchToDevice();
676 m_edgecent[idim][mfi].prefetchToDevice();
681 gshop.updateIntercept({xip,yip},
682 {facetype[1].const_array(), facetype[0].const_array()},
686 nmc = build_faces(vbx, cfg, ftx, fty, lst, xip, yip,
apx,
apy,
fcx,
fcy,
687 dx, problo, cover_multiple_cuts, nsm);
689 build_cells(vbx, cfg, ftx, fty,
apx,
apy, dx, vfr, ctr,
690 bar, bct, bnm, lst, small_volfrac, geom, extend_domain_face,
699 if (nsmallcells == 0 && nmulticuts == 0) {
702 auto ls = m_mgf.getLevelSet();
707 amrex::Print() <<
"AMReX EB: Iter. " << iter+1 <<
" fixed " << nsmallcells
708 <<
" small cells" <<
'\n';
711 amrex::Print() <<
"AMReX EB: Iter. " << iter+1 <<
" fixed " << nmulticuts
712 <<
" multicuts" <<
'\n';
721#pragma omp parallel if (Gpu::notInLaunchRegion())
725 auto& gfab = m_mgf[mfi];
726 auto const&
levelset = gfab.getLevelSet();
732#
if (AMREX_SPACEDIM == 3)
733 auto const& edgetype = gfab.getEdgeType();
738 intercept_to_edge_centroid(xip, yip, zip, xdg, ydg, zdg, clst, dx, problo);
740#elif (AMREX_SPACEDIM == 2)
741 auto& facetype = gfab.getFaceType();
745 intercept_to_edge_centroid(xip, yip, fty, ftx, clst, dx, problo);
749 m_levelset = m_mgf.getLevelSet();
754#if (AMREX_SPACEDIM == 3)
756template <
typename GS>
757requires (std::same_as<GS,STLtools>)
760 int max_grid_size,
int ngrow,
bool extend_domain_face,
int num_crse_opt)
764 prepare_grids(gshop, geom, max_grid_size, ngrow, extend_domain_face, num_crse_opt);
766 if (m_allregular) {
return; }
771 gshop.fillSignedDistance(m_sdf, m_sdf.nGrowVect(), geom);
774 m_marching_cubes[mfi.index()] = std::make_unique<MC::MCFab>();
779#if defined(AMREX_USE_OMP) && !defined(AMREX_USE_GPU)
809 const int coarse_ratio = 2;
810 const int min_width = 8;
811 bool coarsenable = fine_grids.
coarsenable(coarse_ratio, min_width)
812 && (fine_covered_grids.
empty() || fine_covered_grids.
coarsenable(coarse_ratio));
839 AMREX_ASSERT(face_dir >= 0 && face_dir < AMREX_SPACEDIM);
842 if (!m_fc_data[face_dir]) {
843 m_fc_data[face_dir] = std::make_unique<FCData>();
846 if (m_fc_data[face_dir]->m_built) {
return; }
853 m_fc_data[face_dir]->m_built =
true;
861 Level const* refined_cc_level =
nullptr;
862 std::unique_ptr<GShopLevel<G>> transient_level;
867 transient_level = std::make_unique<GShopLevel<G>>(
868 m_parent, gshop, refined_geom,
869 max_grid_size, m_ngrow[0] > 0 ? m_ngrow[0] : 4,
true, 0);
871 if (!transient_level->isOK()) {
872 amrex::Abort(
"GShopLevel::buildFCData: failed to build refined level");
874 refined_cc_level = transient_level.get();
891 m_fc_data[face_dir]->m_covered_grids_fc =
899 auto const& f_levelset = refined_cc_level->
m_levelset;
908#pragma omp parallel reduction(max:mvmc_error)
913 auto const&
fine = f_levelset.const_array(mfi);
917 [=,&tile_error] (
int i,
int j,
int k)
noexcept
919 int ierror = check_mvmc(i,j,k,
fine);
920 tile_error = std::max(tile_error,ierror);
923 mvmc_error = std::max(mvmc_error, tile_error);
930 using ReduceTuple =
typename decltype(reduce_data)::Type;
934 const Box& ndbx = mfi.validbox();
936 auto const&
fine = f_levelset.const_array(mfi);
937 reduce_op.
eval(ndbx, reduce_data,
942 ierror = check_mvmc(i,j,k,fine);
949 ReduceTuple rv = reduce_data.
value(reduce_op);
950 mvmc_error =
amrex::max(0, amrex::get<0>(rv));
967 auto& fc_cellflag = m_fc_data[face_dir]->m_cellflag_fc;
968 MultiFab& fc_volfrac = m_fc_data[face_dir]->m_volfrac_fc;
969 MultiFab& fc_centroid = m_fc_data[face_dir]->m_centroid_fc;
970 MultiFab& fc_bndryarea = m_fc_data[face_dir]->m_bndryarea_fc;
971 MultiFab& fc_bndrycent = m_fc_data[face_dir]->m_bndrycent_fc;
972 MultiFab& fc_bndrynorm = m_fc_data[face_dir]->m_bndrynorm_fc;
973 auto& fc_areafrac = m_fc_data[face_dir]->m_areafrac_fc;
974 auto& fc_facecent = m_fc_data[face_dir]->m_facecent_fc;
975 auto& fc_edgecent = m_fc_data[face_dir]->m_edgecent_fc;
1006 fc_cellflag.define(fc_grids, fc_dmap, 1, ng);
1007 fc_volfrac.
define(fc_grids, fc_dmap, 1, ng);
1008 fc_centroid.
define(fc_grids, fc_dmap, AMREX_SPACEDIM, ng);
1009 fc_bndryarea.
define(fc_grids, fc_dmap, 1, ng);
1010 fc_bndrycent.
define(fc_grids, fc_dmap, AMREX_SPACEDIM, ng);
1011 fc_bndrynorm.
define(fc_grids, fc_dmap, AMREX_SPACEDIM, ng);
1013 for (
int idim = 0; idim < AMREX_SPACEDIM; ++idim) {
1014 fc_areafrac[idim].define(fc_grids, fc_dmap, 1, ng);
1015 fc_facecent[idim].define(fc_grids, fc_dmap, AMREX_SPACEDIM-1, ng);
1025 auto const& cc_cellflag = refined_cc_level->
m_cellflag;
1031 auto const& cc_areafrac = refined_cc_level->
m_areafrac;
1032 auto const& cc_facecent = refined_cc_level->
m_facecent;
1033 auto const& cc_edgecent = refined_cc_level->
m_edgecent;
1040#pragma omp parallel reduction(max:error)
1045 auto const& cvol = fc_volfrac.
array(mfi);
1046 auto const& ccent = fc_centroid.
array(mfi);
1047 auto const& cba = fc_bndryarea.
array(mfi);
1048 auto const& cbc = fc_bndrycent.
array(mfi);
1049 auto const& cbn = fc_bndrynorm.
array(mfi);
1050 AMREX_D_TERM(
auto const& capx = fc_areafrac[0].array(mfi);,
1051 auto const& capy = fc_areafrac[1].array(mfi);,
1052 auto const& capz = fc_areafrac[2].array(mfi););
1053 AMREX_D_TERM(
auto const& cfcx = fc_facecent[0].array(mfi);,
1054 auto const& cfcy = fc_facecent[1].array(mfi);,
1055 auto const& cfcz = fc_facecent[2].array(mfi););
1056 AMREX_D_TERM(
auto const& cecx = fc_edgecent[0].array(mfi);,
1057 auto const& cecy = fc_edgecent[1].array(mfi);,
1058 auto const& cecz = fc_edgecent[2].array(mfi););
1059 auto const& cflag = fc_cellflag.array(mfi);
1067 AMREX_D_TERM(
auto const& fapx = cc_areafrac[0].const_array(mfi);,
1068 auto const& fapy = cc_areafrac[1].const_array(mfi);,
1069 auto const& fapz = cc_areafrac[2].const_array(mfi););
1070 AMREX_D_TERM(
auto const& ffcx = cc_facecent[0].const_array(mfi);,
1071 auto const& ffcy = cc_facecent[1].const_array(mfi);,
1072 auto const& ffcz = cc_facecent[2].const_array(mfi););
1073 AMREX_D_TERM(
auto const& fecx = cc_edgecent[0].const_array(mfi);,
1074 auto const& fecy = cc_edgecent[1].const_array(mfi);,
1075 auto const& fecz = cc_edgecent[2].const_array(mfi););
1076 auto const& fflag = cc_cellflag.const_array(mfi);
1078 Box const& bx = mfi.validbox();
1079 Box const& ndgbx = mfi.grownnodaltilebox(-1,2);
1083 [=,&tile_error] (
int i,
int j,
int k)
noexcept
1086 int ierr = coarsen_from_fine(
AMREX_D_DECL(i,j,k), bx, 2,
1087 cvol,ccent,cba,cbc,cbn,
1091 cflag,fvol,fcent,fba,fbc,fbn,
1096 tile_error = std::max(tile_error,ierr);
1099 error = std::max(error,tile_error);
1106 using ReduceTuple =
typename decltype(reduce_data)::Type;
1110 auto const& cvol = fc_volfrac.
array(mfi);
1111 auto const& ccent = fc_centroid.
array(mfi);
1112 auto const& cba = fc_bndryarea.
array(mfi);
1113 auto const& cbc = fc_bndrycent.
array(mfi);
1114 auto const& cbn = fc_bndrynorm.
array(mfi);
1115 AMREX_D_TERM(
auto const& capx = fc_areafrac[0].array(mfi);,
1116 auto const& capy = fc_areafrac[1].array(mfi);,
1117 auto const& capz = fc_areafrac[2].array(mfi););
1118 AMREX_D_TERM(
auto const& cfcx = fc_facecent[0].array(mfi);,
1119 auto const& cfcy = fc_facecent[1].array(mfi);,
1120 auto const& cfcz = fc_facecent[2].array(mfi););
1123 AMREX_D_TERM(
auto const& cecx = fc_edgecent[0].arrays();,
1124 auto const& cecy = fc_edgecent[1].arrays();,
1125 auto const& cecz = fc_edgecent[2].arrays(););
1126 auto const& cflag = fc_cellflag.array(mfi);
1133 AMREX_D_TERM(
auto const& fapx = cc_areafrac[0].const_array(mfi);,
1134 auto const& fapy = cc_areafrac[1].const_array(mfi);,
1135 auto const& fapz = cc_areafrac[2].const_array(mfi););
1136 AMREX_D_TERM(
auto const& ffcx = cc_facecent[0].const_array(mfi);,
1137 auto const& ffcy = cc_facecent[1].const_array(mfi);,
1138 auto const& ffcz = cc_facecent[2].const_array(mfi););
1139 AMREX_D_TERM(
auto const& fecx = cc_edgecent[0].const_arrays();,
1140 auto const& fecy = cc_edgecent[1].const_arrays();,
1141 auto const& fecz = cc_edgecent[2].const_arrays(););
1142 auto const& fflag = cc_cellflag.const_array(mfi);
1144 const int lidx = mfi.LocalIndex();
1145 Box const& bx = mfi.validbox();
1149 reduce_op.
eval(ndgbx, reduce_data,
1153 int ierr = coarsen_from_fine(
AMREX_D_DECL(i,j,k), bx, 2,
1154 cvol,ccent,cba,cbc,cbn,
1158 cflag,fvol,fcent,fba,fbc,fbn,
1167 ReduceTuple rv = reduce_data.
value(reduce_op);
1183 m_fc_data[face_dir]->m_built =
true;
Fixed-size array types for use on GPU and CPU.
#define BL_PROFILE(a)
Definition AMReX_BLProfiler.H:551
#define AMREX_ALWAYS_ASSERT_WITH_MESSAGE(EX, MSG)
Definition AMReX_BLassert.H:49
#define AMREX_ASSERT(EX)
Definition AMReX_BLassert.H:38
Problem-domain geometry: maps between index space and physical space.
#define AMREX_GPU_DEVICE
Definition AMReX_GpuQualifiers.H:18
amrex::ParmParse pp
Input file parser instance for the given namespace.
Definition AMReX_HypreIJIface.cpp:15
Array4< Real > fine
Definition AMReX_InterpFaceRegister.cpp:90
#define AMREX_D_TERM(a, b, c)
Definition AMReX_SPACE.H:172
#define AMREX_D_DECL(a, b, c)
Definition AMReX_SPACE.H:171
void resize(const Box &b, int N=1, Arena *ar=nullptr)
This function resizes a BaseFab so it covers the Box b with N components.
Definition AMReX_BaseFab.H:1767
Array4< T const > array() const noexcept
Create an Array4 view over all components.
Definition AMReX_BaseFab.H:475
Elixir elixir() noexcept
Transfer owned storage to an Elixir when running in a GPU launch region.
Definition AMReX_BaseFab.H:1810
Reference-counted collection of Boxes.
Definition AMReX_BoxArray.H:676
bool coarsenable(int refinement_ratio, int min_width=1) const
Check whether every Box is coarsenable by refinement_ratio.
Definition AMReX_BoxArray.cpp:615
bool empty() const noexcept
Return whether the BoxArray is empty.
Definition AMReX_BoxArray.H:759
__host__ __device__ BoxND< new_dim > resize() const noexcept
Return a new BoxND of size new_dim by either shrinking or expanding this BoxND.
Definition AMReX_Box.H:927
__host__ __device__ bool contains(const IntVectND< dim > &p) const noexcept
Return true if argument is contained within BoxND.
Definition AMReX_Box.H:233
GpuArray< Real, 3 > CellSizeArray() const noexcept
Returns the cell sizes as a GpuArray for use on host or device.
Definition AMReX_CoordSys.H:85
Calculates the distribution of FABs to MPI processes.
Definition AMReX_DistributionMapping.H:51
static constexpr int ng
Definition AMReX_EB2_MultiGFab.H:26
Definition AMReX_EB2_Level.H:288
GShopLevel(IndexSpace const *is, const Geometry &geom)
Construct an empty level (usually regular) bound to geom.
Definition AMReX_EB2_Level.H:340
void define_fine(G const &gshop, const Geometry &geom, int max_grid_size, int ngrow, bool extend_domain_face, int num_crse_opt)
Define data for the finest level using the supplied GeometryShop.
Definition AMReX_EB2_Level.H:475
void buildFCData(G const &gshop, int face_dir, int max_grid_size)
Build face-centered EB data for face_dir. Builds refined CC transiently.
Definition AMReX_EB2_Level.H:835
GShopLevel(IndexSpace const *is, int ilev, int max_grid_size, int ngrow, const Geometry &geom, GShopLevel< G > &fineLevel)
Construct a coarse level by coarsening fineLevel.
Definition AMReX_EB2_Level.H:794
GShopLevel(IndexSpace const *is, G const &gshop, const Geometry &geom, int max_grid_size, int ngrow, bool extend_domain_face, int num_crse_opt)
Build a level directly from a GeometryShop object.
Definition AMReX_EB2_Level.H:345
static GShopLevel< G > makeAllRegular(IndexSpace const *is, const Geometry &geom)
Build a regular (no EB) level.
Definition AMReX_EB2_Level.H:320
void define_fine_mvmc(GS const &gshop, const Geometry &geom, int max_grid_size, int ngrow, bool extend_domain_face, int num_crse_opt)
Define a fine level for multi-valued multi-cut STL geometries (3-D only)
Definition AMReX_EB2_Level.H:759
Definition AMReX_EB2.H:28
Definition AMReX_EB2_Level.H:42
MultiFab m_bndryarea
Definition AMReX_EB2_Level.H:258
bool isOK() const noexcept
Definition AMReX_EB2_Level.H:48
void fillFaceCent(Array< MultiCutFab *, 3 > const &facecent, const Geometry &geom) const
Populate face centroids for each direction.
Definition AMReX_EB2_Level.cpp:794
MultiFab m_volfrac
Definition AMReX_EB2_Level.H:256
MultiGFab m_mgf
Definition AMReX_EB2_Level.H:253
bool m_allregular
Definition AMReX_EB2_Level.H:265
IntVect m_shift
Definition AMReX_EB2_Level.H:268
iMultiFab m_cutcellmask
Definition AMReX_EB2_Level.H:264
Array< MultiFab, 3 > m_facecent
Definition AMReX_EB2_Level.H:262
void fillBndryAreaFC(MultiFab &bndryarea, int face_dir, const Geometry &geom) const
Fill face-centered boundary areas for face_dir.
Definition AMReX_EB2_Level.cpp:1184
Array< MultiFab, 3 > m_edgecent
Definition AMReX_EB2_Level.H:263
void fillBndryNormFC(MultiFab &bndrynorm, int face_dir, const Geometry &geom) const
Fill face-centered boundary normals for face_dir.
Definition AMReX_EB2_Level.cpp:1220
void prepareForCoarsening(const Level &rhs, int max_grid_size, IntVect const &ngrow)
Prepare for coarsening by copying from rhs, respecting max grid size and grow vector.
Definition AMReX_EB2_Level.cpp:10
void fillCutCellMask(iMultiFab &cutcellmask, const Geometry &geom) const
Populate the cut-cell mask into cutcellmask.
Definition AMReX_EB2_Level.cpp:946
MultiFab m_centroid
Definition AMReX_EB2_Level.H:257
FabArray< EBCellFlagFab > m_cellflag
Definition AMReX_EB2_Level.H:255
const DistributionMapping & DistributionMap() const noexcept
Definition AMReX_EB2_Level.H:110
BoxArray m_covered_grids
Definition AMReX_EB2_Level.H:247
void fillLevelSet(MultiFab &levelset, const Geometry &geom) const
Write the implicit function values into levelset.
Definition AMReX_EB2_Level.cpp:907
Geometry m_geom
Definition AMReX_EB2_Level.H:244
Array< MultiFab, 3 > m_areafrac
Definition AMReX_EB2_Level.H:261
IntVect m_ngrow
Definition AMReX_EB2_Level.H:245
void fillEBCellFlagFC(FabArray< EBCellFlagFab > &cellflag, int face_dir, const Geometry &geom) const
Fill face-centered EBCellFlags for face_dir.
Definition AMReX_EB2_Level.cpp:1238
IndexSpace const * m_parent
Definition AMReX_EB2_Level.H:269
const FCData & getFCData(int face_dir) const
Get reference to face-centered data for face_dir (0=x, 1=y, 2=z).
Definition AMReX_EB2_Level.H:197
BoxArray m_grids
Definition AMReX_EB2_Level.H:246
void fillEdgeCentFC(Array< MultiFab *, 3 > const &edgecent, int face_dir, const Geometry &geom) const
Fill face-centered edge centroids for face_dir.
Definition AMReX_EB2_Level.cpp:1322
const Geometry & Geom() const noexcept
Definition AMReX_EB2_Level.H:120
std::map< int, std::unique_ptr< MC::MCFab > > m_marching_cubes
Definition AMReX_EB2_Level.H:251
bool hasEBInfo() const noexcept
Definition AMReX_EB2_Level.H:137
void write_to_chkpt_file(const std::string &fname, bool extend_domain_face, int max_grid_size) const
Write this level’s EB data to a checkpoint file fname.
Definition AMReX_EB2_Level.cpp:956
void fillEBCellFlag(FabArray< EBCellFlagFab > &cellflag, const Geometry &geom) const
Fill cellflag with EBCellFlag data using geometry geom.
Definition AMReX_EB2_Level.cpp:429
void fillEdgeCent(Array< MultiCutFab *, 3 > const &edgecent, const Geometry &geom) const
Populate edge centroids for each direction.
Definition AMReX_EB2_Level.cpp:835
int coarsenFromFine(Level &fineLevel, bool fill_boundary)
Coarsen EB data from fineLevel into this level.
Definition AMReX_EB2_Level.cpp:84
MultiFab m_bndrynorm
Definition AMReX_EB2_Level.H:260
bool m_has_eb_info
Definition AMReX_EB2_Level.H:267
void fillAreaFrac(Array< MultiCutFab *, 3 > const &areafrac, const Geometry &geom) const
Populate face area fractions for each direction.
Definition AMReX_EB2_Level.cpp:640
MultiFab m_bndrycent
Definition AMReX_EB2_Level.H:259
MultiFab m_levelset
Definition AMReX_EB2_Level.H:254
void setShift(int direction, int ncells)
Shift this level by ncells cells along direction direction (for periodic tiling).
Definition AMReX_EB2_Level.cpp:1047
bool isAllRegular() const noexcept
Definition AMReX_EB2_Level.H:46
static IntVect fcEdgeType(int idim, int face_dir) noexcept
Index type of the face-centered edge centroids for edge direction idim.
Definition AMReX_EB2_Level.H:184
void fillBndryCent(MultiCutFab &bndrycent, const Geometry &geom) const
Populate boundary centroids into bndrycent.
Definition AMReX_EB2_Level.cpp:592
void fillCentroidFC(MultiFab ¢roid, int face_dir, const Geometry &geom) const
Fill face-centered centroids for face_dir.
Definition AMReX_EB2_Level.cpp:1166
void fillBndryNorm(MultiCutFab &bndrynorm, const Geometry &geom) const
Populate boundary normals into bndrynorm.
Definition AMReX_EB2_Level.cpp:616
void fillBndryCentFC(MultiFab &bndrycent, int face_dir, const Geometry &geom) const
Fill face-centered boundary centroids for face_dir.
Definition AMReX_EB2_Level.cpp:1202
MultiFab m_sdf
Definition AMReX_EB2_Level.H:249
IntVect const & nGrowVect() const noexcept
Definition AMReX_EB2_Level.H:125
void fillAreaFracFC(Array< MultiFab *, 3 > const &areafrac, int face_dir, const Geometry &geom) const
Fill face-centered area fractions for face_dir.
Definition AMReX_EB2_Level.cpp:1101
void fillVolFrac(MultiFab &vfrac, const Geometry &geom) const
Populate volume fractions into vfrac using geom.
Definition AMReX_EB2_Level.cpp:483
const BoxArray & boxArray() const noexcept
Definition AMReX_EB2_Level.H:108
Level(IndexSpace const *is, const Geometry &geom)
Construct a Level bound to EB index space is and Geometry geom.
Definition AMReX_EB2_Level.H:113
friend class GShopLevel
Definition AMReX_EB2_Level.H:275
bool m_ok
Definition AMReX_EB2_Level.H:266
void fillCentroid(MultiCutFab ¢roid, const Geometry &geom) const
Populate cut-cell centroids into a MultiCutFab.
Definition AMReX_EB2_Level.cpp:544
Array< std::unique_ptr< FCData >, 3 > m_fc_data
Face-centered data storage: [0]=x-face, [1]=y-face, [2]=z-face.
Definition AMReX_EB2_Level.H:272
DistributionMapping m_dmap
Definition AMReX_EB2_Level.H:248
void fillBndryArea(MultiCutFab &bndryarea, const Geometry &geom) const
Populate boundary-face areas into bndryarea.
Definition AMReX_EB2_Level.cpp:568
void buildCellFlag()
Build EBCellFlag data from the current implicit function field.
Definition AMReX_EB2_Level.cpp:403
void buildCutCellMask(Level const &fine_level)
Build the cut-cell mask by coarsening from fine_level.
Definition AMReX_EB2_Level.cpp:966
bool hasFCData(int face_dir) const noexcept
Check if face-centered data is available for face_dir (0=x, 1=y, 2=z).
Definition AMReX_EB2_Level.H:191
IndexSpace const * getEBIndexSpace() const noexcept
Definition AMReX_EB2_Level.H:122
void fillVolFracFC(MultiFab &vfrac, int face_dir, const Geometry &geom) const
Fill face-centered volume fractions for face_dir.
Definition AMReX_EB2_Level.cpp:1062
void fillFaceCentFC(Array< MultiFab *, 3 > const &facecent, int face_dir, const Geometry &geom) const
Fill face-centered face centroids for face_dir.
Definition AMReX_EB2_Level.cpp:1290
LayoutData wrapper that allocates a GFab per FAB/Box.
Definition AMReX_EB2_MultiGFab.H:86
FAB storing EBCellFlag data per cell.
Definition AMReX_EBCellFlag.H:324
An Array of FortranArrayBox(FAB)-like Objects.
Definition AMReX_FabArray.H:356
Array4< typename FabArray< FAB >::value_type const > array(const MFIter &mfi) const noexcept
Read-only Array4 view for the FAB referenced by iterator mfi.
Definition AMReX_FabArray.H:621
Array4< typename FabArray< FAB >::value_type const > const_array(const MFIter &mfi) const noexcept
Synonym for array(const MFIter&) that highlights read-only semantics.
Definition AMReX_FabArray.H:649
Rectangular problem domain geometry.
Definition AMReX_Geometry.H:85
const Box & Domain() const noexcept
Returns our rectangular domain.
Definition AMReX_Geometry.H:244
Periodicity periodicity() const noexcept
Return the Periodicity based on the length of the domain.
Definition AMReX_Geometry.H:415
GpuArray< Real, 3 > ProbLoArray() const noexcept
Return the lo end of the problem domain in a GpuArray for device code.
Definition AMReX_Geometry.H:217
bool isPeriodic(int dir) const noexcept
Is the domain periodic in the specified direction?
Definition AMReX_Geometry.H:388
Definition AMReX_GpuElixir.H:13
__host__ static __device__ constexpr IntVectND< dim > TheCellVector() noexcept
This static member function returns a reference to a constant IntVectND object, all of whose dim argu...
Definition AMReX_IntVect.H:810
__host__ static __device__ constexpr IntVectND< dim > TheZeroVector() noexcept
This static member function returns a reference to a constant IntVectND object, all of whose dim argu...
Definition AMReX_IntVect.H:771
__host__ static __device__ constexpr IntVectND< dim > TheNodeVector() noexcept
This static member function returns a reference to a constant IntVectND object, all of whose dim argu...
Definition AMReX_IntVect.H:801
__host__ static __device__ constexpr IntVectND< dim > TheDimensionVector(int d) noexcept
This static member function returns a reference to a constant IntVectND object, all of whose dim argu...
Definition AMReX_IntVect.H:790
Iterator for looping ever tiles and boxes of amrex::FabArray based containers.
Definition AMReX_MFIter.H:88
bool isValid() const noexcept
Is the iterator valid i.e. is it associated with a FAB?
Definition AMReX_MFIter.H:172
FabArray façade that only allocates cut-cell Fabs (skipping regular cells).
Definition AMReX_MultiCutFab.H:87
A collection (stored as an array) of FArrayBox objects.
Definition AMReX_MultiFab.H:40
void define(const BoxArray &bxs, const DistributionMapping &dm, int nvar, int ngrow, const MFInfo &info=MFInfo(), const FabFactory< FArrayBox > &factory=FArrayBoxFactory())
Definition AMReX_MultiFab.cpp:519
Parse Parameters From Command Line and Input Files.
Definition AMReX_ParmParse.H:351
int queryAdd(std::string_view name, T &ref)
If name is found, the value in the ParmParse database will be stored in the ref argument....
Definition AMReX_ParmParse.H:1047
This class provides the user with a few print options.
Definition AMReX_Print.H:35
Definition AMReX_Reduce.H:438
Type value()
Definition AMReX_Reduce.H:473
Definition AMReX_Reduce.H:597
void eval(MF const &mf, IntVect const &nghost, D &reduce_data, F &&f)
Definition AMReX_Reduce.H:731
A Collection of IArrayBoxes.
Definition AMReX_iMultiFab.H:34
amrex_real Real
Floating Point Type for Fields.
Definition AMReX_REAL.H:79
amrex_long Long
Definition AMReX_INT.H:30
__host__ __device__ BoxND< dim > convert(const BoxND< dim > &b, const IntVectND< dim > &typ) noexcept
Return a copy of b converted to the nodal flags typ.
Definition AMReX_Box.H:1630
__host__ __device__ BoxND< dim > surroundingNodes(const BoxND< dim > &b, int dir) noexcept
Return a BoxND with NODE based coordinates in direction dir that encloses BoxND b.
Definition AMReX_Box.H:1582
__host__ __device__ BoxND< dim > coarsen(const BoxND< dim > &b, int ref_ratio) noexcept
Return a copy of b coarsened by the isotropic ratio ref_ratio.
Definition AMReX_Box.H:1469
__host__ __device__ BoxND< dim > enclosedCells(const BoxND< dim > &b, int dir) noexcept
Return a BoxND with CELL based coordinates in direction dir that is enclosed by b.
Definition AMReX_Box.H:1664
__host__ __device__ BoxND< dim > grow(const BoxND< dim > &b, int i) noexcept
Return a copy of b grown uniformly by i cells in every direction.
Definition AMReX_Box.H:1326
__host__ __device__ BoxND< dim > refine(const BoxND< dim > &b, int ref_ratio) noexcept
Return a copy of b refined by the isotropic ratio ref_ratio.
Definition AMReX_Box.H:1510
std::array< T, N > Array
Definition AMReX_Array.H:31
int MyProc() noexcept
Definition AMReX_ParallelDescriptor.H:128
void ReduceBoolOr(bool &)
Definition AMReX_ParallelDescriptor.cpp:1292
int NProcs() noexcept
Definition AMReX_ParallelDescriptor.H:255
__host__ __device__ constexpr const T & max(const T &a, const T &b) noexcept
Definition AMReX_Algorithm.H:53
Definition AMReX_FabArrayBase.H:38
void streamSynchronize() noexcept
Definition AMReX_GpuDevice.H:310
bool inLaunchRegion() noexcept
Definition AMReX_GpuControl.H:88
bool notInLaunchRegion() noexcept
Definition AMReX_GpuControl.H:89
void write_stl(std::string const &filename, std::map< int, std::unique_ptr< MCFab > > const &mc_fabs)
Write the collected marching-cubes output to an STL file.
Definition AMReX_MarchingCubes.cpp:924
void marching_cubes(Geometry const &geom, FArrayBox &sdf_fab, MCFab &mc_fab)
Run marching cubes on signed-distance field sdf_fab.
Definition AMReX_MarchingCubes.cpp:693
void Initialize()
Initialize internal lookup tables and device buffers for marching cubes.
Definition AMReX_MarchingCubes.cpp:651
MPI_Comm CommunicatorSub() noexcept
sub-communicator for current frame
Definition AMReX_ParallelContext.H:70
Definition AMReX_Amr.cpp:50
__host__ __device__ void ignore_unused(const Ts &...)
No-op helper that marks variables as intentionally unused.
Definition AMReX.H:259
RunOn
Definition AMReX_GpuControl.H:65
void AllGatherBoxes(Vector< Box > &bxs, int n_extra_reserve)
Gather boxes from all MPI ranks into bxs.
Definition AMReX_Box.cpp:124
BoxND< 3 > Box
Box is an alias for amrex::BoxND instantiated with AMREX_SPACEDIM.
Definition AMReX_BaseFwd.H:35
IntVectND< 3 > IntVect
IntVect is an alias for amrex::IntVectND instantiated with AMREX_SPACEDIM.
Definition AMReX_BaseFwd.H:38
__host__ __device__ constexpr IntVectND< dim > scale(const IntVectND< dim > &p, int s) noexcept
Returns a IntVectND obtained by multiplying each of the components of this IntVectND by s.
Definition AMReX_IntVect.H:1102
int Verbose() noexcept
Return the verbosity level configured via ParmParse or SetVerbose().
Definition AMReX.cpp:182
void Abort(const std::string &msg)
Print a fatal-error message to stderr and abort execution.
Definition AMReX.cpp:242
void LoopOnCpu(Dim3 lo, Dim3 hi, F const &f) noexcept
Definition AMReX_Loop.H:365
A multidimensional array accessor.
Definition AMReX_Array4.H:288
Face-centered EB data storage structure.
Definition AMReX_EB2_Level.H:149
bool m_built
Definition AMReX_EB2_Level.H:173
Array< MultiFab, 3 > m_areafrac_fc
Definition AMReX_EB2_Level.H:156
MultiFab m_bndryarea_fc
Definition AMReX_EB2_Level.H:153
Array< MultiFab, 3 > m_edgecent_fc
Definition AMReX_EB2_Level.H:158
BoxArray m_covered_grids_fc
Definition AMReX_EB2_Level.H:172
MultiFab m_bndrycent_fc
Definition AMReX_EB2_Level.H:154
MultiFab m_centroid_fc
Definition AMReX_EB2_Level.H:152
MultiFab m_bndrynorm_fc
Definition AMReX_EB2_Level.H:155
MultiFab m_volfrac_fc
Definition AMReX_EB2_Level.H:151
Array< MultiFab, 3 > m_facecent_fc
Definition AMReX_EB2_Level.H:157
FabArray< EBCellFlagFab > m_cellflag_fc
Definition AMReX_EB2_Level.H:150
FabArray memory allocation information.
Definition AMReX_FabArray.H:73
MFInfo & SetTag() noexcept
Terminate a variadic SetTag() call.
Definition AMReX_FabArray.H:90
MFInfo & SetAlloc(bool a) noexcept
Control whether FAB storage is allocated when the FabArray is defined.
Definition AMReX_FabArray.H:81
Definition AMReX_MFIter.H:20
MFItInfo & SetDynamic(bool f) noexcept
Definition AMReX_MFIter.H:43